# Purple::HID.pm
#
# Eugene Eric Kim <eekim@eekim.com>
# http://www.eekim.com/software/purple/
#
# $Id: HID.pm,v 1.3 2001/05/01 18:26:32 eekim Exp $
#
# Copyright (c) Eugene Eric Kim 2000-2001.  All rights reserved.
# See COPYING for licensing terms.

package Purple::HID;

=head1 NAME

Purple::HID - class for manipulating Augment-style hierarchical addresses.

=head1 SYNOPSIS

  use Purple::HID;
  my $hid = new Purple::HID;  # 1

  $hid->spawn;  # 1A
  $hid->incr;   # 1B
  $hid->spawn;  # 1B1
  $hid->incr;   # 1B2
  $hid->incr;   # 1B3
  $hid->spawn;  # 1B3A
  $hid->incr;   # 1B3B
  $hid->kill;   # 1B3
  $hid->kill;   # 1B
  $hid->kill;   # 1
  $hid->incr;   # 2

  if ($hid->string eq '2') {  # true!
    $hid->print;  # prints '2'
  }

=head1 DESCRIPTION

Implements Augment-style hierarchical addresses (explained at
http://www.bootstrap.org/augdocs/oad-2250.htm).

=cut

use strict;

### public methods

=head1 METHODS

=head2 new()

 pre:
   none
 post:
   none

Constructor.  Initializes address at '0'.

=cut

sub new {
  my $this = shift;
  my $self = { HID => [1] };
  bless($self, $this);
  return $self;
}

=head2 incr()

 pre:
   none
 post:
   none

Increments the address.  If the address is numerical, just increments
it.  If the address is alphabetical and the current address is 'Z',
adds another letter, starting at 'A'.

=cut

sub incr {
  my $this = shift;
  my $tail = pop @{$this->{HID}};
  if ($tail =~ /[0-9]/) {
    $tail++;
    push @{$this->{HID}}, $tail;
  }
  else {
    my $last_char;
    $tail =~ s/^(.*)(.)$/$1/;
    $last_char = $2;
    if ($last_char eq 'Z') {
      $tail .= 'A';
    }
    else {
      $last_char = chr(ord($last_char) + 1);
      $tail .= $last_char;
    }
    push @{$this->{HID}}, $tail;
  }
}

=head2 spawn()

 pre:
   none
 post:
   none

Spawns a new address, alternating alphabetical with numerical.

=cut

sub spawn {
  my $this = shift;
  my $tail = ${$this->{HID}}[$#{$this->{HID}}];
  if ($tail =~ /[0-9]/) {
    push @{$this->{HID}}, 'A';
  }
  else {
    push @{$this->{HID}}, '1';
  }
}

=head2 kill()

 pre:
   none
 post:
   none

Kills the most recently spawned address.

=cut

sub kill {
  my $this = shift;

  pop @{$this->{HID}};
}

=head2 string()

 pre:
   none
 post:
   $hid_string

Returns a string with the hierarchical address.

=cut

sub string {
  my $this = shift;
  my ($hid, $hid_string);

  foreach $hid (@{$this->{HID}}) {
    $hid_string .= $hid;
  }
  return $hid_string;
}

=head2 print()

 pre:
   none
 post:
   none

Prints the hierarchical address with a newline.

=cut

sub print {
  my $this = shift;

  print $this->string . "\n";
}

1;

=head1 AUTHOR

Eugene Eric Kim <eekim@eekim.com>

=cut
